/* -----------------------------------------------------------------------------
DISCLAIMER:

DeleD .dmf parser is released under the MIT license
Copyright (c) 2005 Ryan Fleet

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the
Software without restriction, including without
limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.

DeleD is Copyright (c) 2005 Delgine Team
-----------------------------------------------------------------------------

REVISION HISTORY:

        june/23/2005: Ryan Fleet (rmfleet@gmail.com)
            - Original Creation
        july/03/2005: Ryan Fleet (rmfleet@gmail.com)
		    - First Release

-----------------------------------------------------------------------------

KNOWN ISSUES / LIMITATIONS:

        july/03/2005: does not support versions previous to 0.91

-----------------------------------------------------------------------------

USAGE EXAMPLE:

#include "dmf.h"

unsigned int TextureLoadCallback(const dmf_layer_t *layer)
{
    unsigned int index = 0;

    // texture loading routines. should return a texture index.

    return index;
}

void ProgressCallback(unsigned int stage, unsigned int progress)
{
    // progress is in the range 0% - 100%

    switch(stage)
    {
    case DMF_LOAD_STAGE_HEADER:
        break;
    case DMF_LOAD_STAGE_MATERIALS:
        break;
    case DMF_LOAD_STAGE_OBJECTS:
        break;
    case DMF_LOAD_STAGE_LIGHTS:
        break;
    };
}

int main(int argc, char *argv[])
{
    dmf_file_t dmf;
    unsigned int result;

    result = LoadDMF(&dmf, "example.dmf", ProgressCallback, TextureLoadCallback);
    if(DMF_NO_ERROR != result)
    {
        // notify of error
        return 0;
    }

    TriangulateDMF(&dmf);

    // do some stuff with the data. ie convert or modify.

    if(dmf.version >= DMF_VERSION_1_1)
        ScaleDMF(&map, 1.f, 2.f, 2.f, 1.f, -1.f);
    else
        ScaleDMF(&map, 1.f, 2.f, 2.f, 1.f, 1.f);

    WriteDMF(&dmf, "example2.dmf");
    FreeDMF(&dmf);

    return 0;
}

The above example will load example.dmf and store the file information in
dmf_file_t dmf; Each call to LoadDMF should be matched by a call to FreeDMF
to release the resources. The last two parameters are optional callback
functions.  ProgressCallback is used to communicate back with the application
about the current loading stage. TextureLoadCallback is used to load an image
file and upload the data to the graphics processor.  It should then return an
index back to the material layer which called the function.  This index should
be a reference to the texture data. Both of these callbacks can safely be set
to 'null' (0) if not required by the application.

The function TriangulateDMF(dmf_file_t *df) will convert all geometry to
triangles if your rendering platform requires it.

WriteDMF(dmf_file_t *df) can be used to write out a dmf file either created
from scratch or from a previously loaded file which has been modified.

ScaleDMF(dmf_file_t *df, float x, float y, float z, float u, float v) can be
used to scale vertex, texture coordinate and light position data.

-----------------------------------------------------------------------------*/

#ifndef dmf_h__
#define dmf_h__

#ifdef __cplusplus
extern "C" {
#endif

#define DMF_MAX_TOKEN_SIZE             2048 /* buffer large enough to hold a token */

#define DMF_FILE_HEADER                "DeleD Map File" /* do not modify */

#define DMF_CHAR_VERSION_0_9_1         "Version 0.91" /* do not modify */
#define DMF_CHAR_VERSION_1_0           "Version 1.0"  /* do not modify */
#define DMF_CHAR_VERSION_1_1           "Version 1.1"  /* do not modify */

#define DMF_VERSION_0_9_1              1
#define DMF_VERSION_1_0                2
#define DMF_VERSION_1_1                3

#define DMF_FALSE                      0
#define DMF_TRUE                       -1

#define DMF_LAYER_TYPE_UNUSED          -1
#define DMF_LAYER_TYPE_TEXTURE         0
#define DMF_LAYER_TYPE_COLOR           1

#define DMF_LIGHT_TYPE_POINT           1
#define DMF_LIGHT_TYPE_DIRECTIONAL     2
#define DMF_LIGHT_TYPE_SPOT            3

#define DMF_LOAD_STAGE_HEADER          1
#define DMF_LOAD_STAGE_OBJECTS         2
#define DMF_LOAD_STAGE_MATERIALS       3
#define DMF_LOAD_STAGE_LIGHTS          4

#define DMF_BLENDOP_INACTIVE           0
#define DMF_BLENDOP_REPLACE            1
#define DMF_BLENDOP_ADD                2
#define DMF_BLENDOP_MODULATE           3
#define DMF_BLENDOP_ALPHABLEND         4

#define DMF_SNAP_NONE                  0
#define DMF_SNAP_VERTICES              1
#define DMF_SNAP_CENTER                2

#define DMF_OBJECT_TYPE_RECTANGLE      0
#define DMF_OBJECT_TYPE_CUBE           1
#define DMF_OBJECT_TYPE_PYRAMID        2
#define DMF_OBJECT_TYPE_CYLINDER       3
#define DMF_OBJECT_TYPE_POLYLINE       4
#define DMF_OBJECT_TYPER_VERTICE       5
#define DMF_OBJECT_TYPE_GEOSPHERE      6
#define DMF_OBJECT_TYPE_TORUS          7
#define DMF_OBJECT_TYPE_PLANE          8
#define DMF_OBJECT_TYPE_USER_DEFINED   9
#define DMF_OBJECT_TYPE_PREFAB         10
#define DMF_OBJECT_TYPE_UNDEFINED      11
#define DMF_OBJECT_TYPE_GRID           12
#define DMF_OBJECT_TYPE_POINTLIGHT     13
#define DMF_OBJECT_TYPE_DIRECTLIGHT    14
#define DMF_OBJECT_TYPE_SPOTLIGHT      15
#define DMF_OBJECT_TYPE_PRIMITIVE      16
#define DMF_OBJECT_TYPE_LIGHT          17
#define DMF_OBJECT_TYPE_ELLIPSE        18
#define DMF_OBJECT_TYPE TEXT           19

#define DMF_NO_ERROR                   0
#define DMF_MEMORY_ERROR               1
#define DMF_FILE_ERROR                 2
#define DMF_TOKEN_ERROR                3
#define DMF_VERSION_ERROR              4
#define DMF_FORMAT_ERROR               5
#define DMF_UNSUPPORTED_ERROR          6

typedef signed int  dmfbool;

typedef struct
{
    unsigned char r, g, b, a;

} dmf_rgba_t;

typedef struct
{
    float x, y;

} dmf_vector2_t;

typedef struct
{
    float x, y, z;

} dmf_vector3_t;

typedef struct
{
    char            *name;
    unsigned char   intensity;
    unsigned char   specular[3];
    unsigned char   diffuse[3];
    unsigned int    id;
    unsigned int    type;
    int             umbra;
    int             penumbra;
    int             falloff;
    dmfbool         active;
    int             flash;
    int             range;
    float           center[3];
    float           quadratic;
    float           linear;
    float           constant;
    float           direction[3];
} dmf_light_t;

typedef struct
{
    int             type;
    int             blendop;
    char            *texture;
    unsigned int    textureIndex;

} dmf_layer_t;

typedef struct
{
    unsigned int    numVertices;
    unsigned int    materialIndex;
    unsigned int    *indices;
    dmf_vector2_t   *texcoords[4];

} dmf_polygon_t;

typedef struct
{
    char            *name;
    char            *category;
    unsigned int    id;
    unsigned int    reserved;
    unsigned int    numLayers;
    dmf_layer_t     *layers;

} dmf_material_t;

typedef struct
{

    char            *name;
    char            *tag;
    int             id;
    int             group;
    int             type;
    int             snap;
    dmfbool         auto_uv;
    dmfbool         visible;
    unsigned int    numPolygons;
    unsigned int    numVertices;
    dmf_polygon_t   *polygons;
    dmf_vector3_t   *vertices;

} dmf_object_t;

typedef struct
{
    char            *mapName;
    unsigned int	glDisplayList;
	unsigned int    numLights;
    unsigned int    numMaterials;
    unsigned int    numObjects;
    unsigned int    shadowOpacity;
    unsigned int    version;
    dmf_rgba_t      ambientLight;
    dmf_material_t  *materialList;
    dmf_object_t    *objectList;
    dmf_light_t     *lightList;

} dmf_file_t;

/* callback types */
typedef void (*dmf_progress_cb)(unsigned int stage, unsigned int progress);
typedef unsigned int (*dmf_texture_cb)(const dmf_layer_t *layer);

/* interface */
extern unsigned int FreeDMF(dmf_file_t *df);
extern unsigned int ReadDMF(dmf_file_t *df, const char *filename, dmf_progress_cb pcallback, dmf_texture_cb tcallback);
extern unsigned int ScaleDMF(dmf_file_t *df, float x, float y, float z, float u, float v);
extern unsigned int TriangulateDMF(dmf_file_t *df);
extern unsigned int WriteDMF(const dmf_file_t *df, const char *filename);

#ifdef __cplusplus
}
#endif

#endif /* dmf_h__ */

/* end of file */

