// Model.c


#include "Model.h"
#include <GL/glext.h>

extern int 			mojoTexManLoad( char *filename, int tex_id, int mipmap, GLint flag  );
extern int 			mojoTexManGetFreeId( void );


void Model_New(MODEL * mod)
{
	mod->numMeshes 		= 0;
	mod->pMeshes 		= NULL;
	mod->numMaterials 	= 0;
	mod->pMaterials 	= NULL;
	mod->numTriangles 	= 0;
	mod->pTriangles 	= NULL;
	mod->numVertices 	= 0;
	mod->pVertices 		= NULL;
	mod->numJoints 		= 0;
	mod->pJoints 		= NULL;
	mod->pTimer 		= Timer_GetNewTimer();
	mod->looping 		= TRUE;
	mod->pModelName		= NULL;
}

void Model_Delete(MODEL * mod)
{
	int i;
	for ( i = 0; i < mod->numMeshes; i++ )
		free(mod->pMeshes[i].pTriangleIndices);

	for ( i = 0; i < mod->numMaterials; i++ )
		free(mod->pMaterials[i].pTextureFilename);

	mod->numMeshes = 0;
	if ( mod->pMeshes != NULL )
	{
		free(mod->pMeshes);
		mod->pMeshes = NULL;
	}

	mod->numMaterials = 0;
	if ( mod->pMaterials != NULL )
	{
		free(mod->pMaterials);
		mod->pMaterials = NULL;
	}

	mod->numTriangles = 0;
	if ( mod->pTriangles != NULL )
	{
		free(mod->pTriangles);
		mod->pTriangles = NULL;
	}

	mod->numVertices = 0;
	if ( mod->pVertices != NULL )
	{
		free(mod->pVertices);
		mod->pVertices = NULL;
	}

	for ( i = 0; i < mod->numJoints; i++ )
	{
		free(mod->pJoints[i].pRotationKeyframes);
		free(mod->pJoints[i].pTranslationKeyframes);
	}

	mod->numJoints = 0;
	if ( mod->pJoints != NULL )
	{
		free(mod->pJoints);
		mod->pJoints = NULL;
	}

	free(mod->pModelName);

	free(mod->pTimer);
}

void Model_Draw(MODEL * mod)
{

	//Model_AdvanceAnimation(mod);

	//GLboolean texEnabled = glIsEnabled( GL_TEXTURE_2D );

	// Draw by group
	for ( int i = 0; i < mod->numMeshes; i++ )
	{
		int materialIndex = mod->pMeshes[i].materialIndex;
		if ( materialIndex >= 0 )
		{
			glMaterialfv( GL_FRONT, GL_AMBIENT, mod->pMaterials[materialIndex].ambient );
			glMaterialfv( GL_FRONT, GL_DIFFUSE, mod->pMaterials[materialIndex].diffuse );
			//glMaterialfv( GL_FRONT, GL_SPECULAR, mod->pMaterials[materialIndex].specular );
			//glMaterialfv( GL_FRONT, GL_EMISSION, mod->pMaterials[materialIndex].emissive );
			//glMaterialf( GL_FRONT, GL_SHININESS, mod->pMaterials[materialIndex].shininess );

			if ( mod->pMaterials[materialIndex].texture > 0 )
			{
				glBindTexture( GL_TEXTURE_2D, mod->pMaterials[materialIndex].texture );
				glEnable( GL_TEXTURE_2D );
			}
			else
				glDisable( GL_TEXTURE_2D );
		}
		else
		{
			// Material properties?
			glDisable( GL_TEXTURE_2D );
		}

		glBegin( GL_TRIANGLES );
		{
			for ( int j = 0; j < mod->pMeshes[i].numTriangles; j++ )
			{
				int triangleIndex = mod->pMeshes[i].pTriangleIndices[j];
				const TRIANGLE * pTri = &mod->pTriangles[triangleIndex];

				for ( int k = 0; k < 3; k++ )
				{
					int index = pTri->vertexIndices[k];

					if ( mod->pVertices[index].boneID == -1 )
					{
						glTexCoord2f( pTri->s[k], pTri->t[k] );
						glNormal3fv( pTri->vertexNormals[k] );
						glVertex3fv( mod->pVertices[index].location );
					}
					else
					{
						// rotate according to transformation MsMatrix
						float * mat_final = mod->pJoints[mod->pVertices[index].boneID].mat_final;

						glTexCoord2f( pTri->s[k], pTri->t[k] );

						float vec_newNormal[4];
						MsVector_MsVector( vec_newNormal, pTri->vertexNormals[k] );
						MsVector_Transform3( vec_newNormal, mat_final );
						MsVector_Normalize( vec_newNormal );
						glNormal3fv( vec_newNormal );

						float vec_newVertex[4];
						MsVector_MsVector( vec_newVertex, mod->pVertices[index].location );
						MsVector_Transform( vec_newVertex, mat_final );
						glVertex3fv( vec_newVertex );
					}
				}
			}
		}
		glEnd();
	}

/*	if ( texEnabled )
		glEnable( GL_TEXTURE_2D );
	else
		glDisable( GL_TEXTURE_2D );
*/
}

void Model_ReloadTextures(MODEL * mod)
{
	int loctex;

	for ( int i = 0; i < mod->numMaterials; i++ )
		if ( strlen( mod->pMaterials[i].pTextureFilename ) > 0 )
		{
			loctex = mojoTexManGetFreeId();
			mojoTexManLoad( mod->pMaterials[i].pTextureFilename, loctex, 1, GL_CLAMP_TO_EDGE  );
			mod->pMaterials[i].texture = loctex;
			printf("loading model %s #texid %d", mod->pMaterials[i].pTextureFilename, loctex);
			//mod->pMaterials[i].texture = LoadGLTexture( mod->pMaterials[i].pTextureFilename );
		}else
			mod->pMaterials[i].texture = 0;
}

void Model_SetJointKeyframe(MODEL * mod, int jointIndex, int keyframeIndex, float time, float *parameter, BOOL isRotation )
{
	//assert( mod->numJoints > jointIndex );

	KEYFRAME * keyframe = isRotation ? &mod->pJoints[jointIndex].pRotationKeyframes[keyframeIndex] :
		&mod->pJoints[jointIndex].pTranslationKeyframes[keyframeIndex];

	keyframe->jointIndex = jointIndex;
	keyframe->time = time;
	memcpy( keyframe->parameter, parameter, sizeof( float )*3 );
}

void Model_SetupJoints(MODEL * mod)
{
	int i;
	for ( i = 0; i < mod->numJoints; i++ )
	{
		JOINT * joint = &mod->pJoints[i];

		MsMatrix_SetRotationRadians( joint->mat_relative, joint->localRotation );
		MsMatrix_SetTranslation( joint->mat_relative, joint->localTranslation );
		if ( joint->parent != -1 )
		{
			MsMatrix_Set( joint->mat_absolute, mod->pJoints[joint->parent].mat_absolute);
			MsMatrix_PostMultiply( joint->mat_absolute, joint->mat_relative );
		}
		else
			MsMatrix_Set( joint->mat_absolute, joint->mat_relative);
	}

	for ( i = 0; i < mod->numVertices; i++ )
	{
		VERTEX * vertex = &mod->pVertices[i];

		if ( vertex->boneID != -1 )
		{
			float * MsMatrix = mod->pJoints[vertex->boneID].mat_absolute;

			MsMatrix_InverseTranslateVect( MsMatrix, vertex->location );
			MsMatrix_InverseRotateVect( MsMatrix, vertex->location );
		}
	}

	for ( i = 0; i < mod->numTriangles; i++ )
	{
		TRIANGLE * triangle = &mod->pTriangles[i];

		for ( int j = 0; j < 3; j++ ) {

			const VERTEX * vertex = &mod->pVertices[triangle->vertexIndices[j]];
			if ( vertex->boneID != -1 ) {
				float * MsMatrix = mod->pJoints[vertex->boneID].mat_absolute;
				MsMatrix_InverseRotateVect( MsMatrix, triangle->vertexNormals[j] );
			}
		}
	}
}

void Model_Restart(MODEL * mod)
{
	for ( int i = 0; i < mod->numJoints; i++ )
	{
		mod->pJoints[i].currentRotationKeyframe = mod->pJoints[i].currentTranslationKeyframe = 0;
		MsMatrix_Set( mod->pJoints[i].mat_final, mod->pJoints[i].mat_absolute);
	}

	Timer_Reset(mod->pTimer);
}

void Model_AdvanceAnimation(MODEL * mod, float start, float end)
{
	float time = start * (1.0/24.0)*1000.0+Timer_GetTime(mod->pTimer);

	if ( time > (end-1) * (1.0/24.0)*1000.0/*mod->totalTime*/ )
	{
		//if ( mod->looping )
		//{
			Model_Restart(mod);
			time = start*(1.0/24.0)*1000.0;
		//}
		//else
		//	time = mod->totalTime;
	}

	//printf("model time #%f\r", time);

	for ( int i = 0; i < mod->numJoints; i++ )
	{
		float transVec[3];
		float mat_transform[16];
		MsMatrix_LoadIdentity(mat_transform);
		int frame;
		JOINT * pJoint = &mod->pJoints[i];

		if ( pJoint->numRotationKeyframes == 0 && pJoint->numTranslationKeyframes == 0 )
		{
			MsMatrix_Set( pJoint->mat_final, pJoint->mat_absolute);
			continue;
		}

		frame = pJoint->currentTranslationKeyframe;



		while ( frame < pJoint->numTranslationKeyframes && pJoint->pTranslationKeyframes[frame].time < time )
		{
			frame++;
		}
		pJoint->currentTranslationKeyframe = frame;

		if ( frame == 0 )
			memcpy( transVec, pJoint->pTranslationKeyframes[0].parameter, sizeof ( float )*3 );
		else if ( frame == pJoint->numTranslationKeyframes )
			memcpy( transVec, pJoint->pTranslationKeyframes[frame-1].parameter, sizeof ( float )*3 );
		else
		{
			//assert( frame > 0 && frame < pJoint->numTranslationKeyframes );

			const KEYFRAME curFrame  = pJoint->pTranslationKeyframes[frame];
			const KEYFRAME prevFrame = pJoint->pTranslationKeyframes[frame-1];

			float timeDelta = curFrame.time-prevFrame.time;
			float interpValue = ( float )(( time-prevFrame.time )/timeDelta );

			transVec[0] = prevFrame.parameter[0]+( curFrame.parameter[0]-prevFrame.parameter[0] )*interpValue;
			transVec[1] = prevFrame.parameter[1]+( curFrame.parameter[1]-prevFrame.parameter[1] )*interpValue;
			transVec[2] = prevFrame.parameter[2]+( curFrame.parameter[2]-prevFrame.parameter[2] )*interpValue;
		}

		frame = pJoint->currentRotationKeyframe;
		while ( frame < pJoint->numRotationKeyframes && pJoint->pRotationKeyframes[frame].time < time )
		{
			frame++;
		}
		pJoint->currentRotationKeyframe = frame;

		if ( frame == 0 )
			MsMatrix_SetRotationRadians( mat_transform, pJoint->pRotationKeyframes[0].parameter );
		else if ( frame == pJoint->numRotationKeyframes )
			MsMatrix_SetRotationRadians( mat_transform, pJoint->pRotationKeyframes[frame-1].parameter );
		else
		{
			//assert( frame > 0 && frame < pJoint->numRotationKeyframes );

			const KEYFRAME curFrame = pJoint->pRotationKeyframes[frame];
			const KEYFRAME prevFrame = pJoint->pRotationKeyframes[frame-1];

			float timeDelta = curFrame.time-prevFrame.time;
			float interpValue = ( float )(( time-prevFrame.time )/timeDelta );

//			assert( interpValue >= 0 && interpValue <= 1 );

			float rotVec[3];

			rotVec[0] = prevFrame.parameter[0]+( curFrame.parameter[0]-prevFrame.parameter[0] )*interpValue;
			rotVec[1] = prevFrame.parameter[1]+( curFrame.parameter[1]-prevFrame.parameter[1] )*interpValue;
			rotVec[2] = prevFrame.parameter[2]+( curFrame.parameter[2]-prevFrame.parameter[2] )*interpValue;

			MsMatrix_SetRotationRadians( mat_transform, rotVec );
		}

		MsMatrix_SetTranslation( mat_transform, transVec );
		float mat_relativeFinal[16];
	    memcpy(mat_relativeFinal, pJoint->mat_relative, sizeof(float)*16);
		MsMatrix_PostMultiply( mat_relativeFinal, mat_transform );

		if ( pJoint->parent == -1 )
			MsMatrix_Set( pJoint->mat_final, mat_relativeFinal);
		else
		{
			MsMatrix_Set( pJoint->mat_final, mod->pJoints[pJoint->parent].mat_final);
			MsMatrix_PostMultiply( pJoint->mat_final, mat_relativeFinal );
		}
	}
}

