#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <GL/gl.h>

#include "mojovector.h"
#include "mojomatrix.h"

__inline void VectorLoad(VECTOR *v, float x, float y, float z) {
  v->v[X] = x;
  v->v[Y] = y;
  v->v[Z] = z;
}

__inline void VectorGet(VECTOR *v, float *x, float *y, float *z) {
  *x = v->v[X];
  *y = v->v[Y];
  *z = v->v[Z];
}

/*
 * Rotate a vector around the X axis
 */
__inline void VectorRotateX(VECTOR *v, float pitch) {
  float a = pitch * 2 * PI / 360.0f;
  float s = (float)sin(a);
  float c = (float)cos(a);
  float y = v->v[Y];
  float z = v->v[Z];

  v->v[Y] = y*c - z*s;
  v->v[Z] = y*s + z*c;
}

__inline void VectorRotateY(VECTOR *v, float yaw) {
  float a = yaw * 2 * PI / 360.0f;
  float s = (float)sin(a);
  float c = (float)cos(a);
  float x = v->v[X];
  float z = v->v[Z];

  v->v[X] = x*c + z*s;
  v->v[Z] = z*c - x*s;
}

__inline void VectorRotateZ(VECTOR *v, float roll) {
  float a = roll * 2 * PI / 360.0f;
  float s = (float)sin(a);
  float c = (float)cos(a);
  float x = v->v[X];
  float y = v->v[Y];

  v->v[X] = x*c - y*s;
  v->v[Y] = y*c + x*s;
}

__inline void VectorCopy(VECTOR *v, VECTOR *w) {
  v->v[X] = w->v[X];
  v->v[Y] = w->v[Y];
  v->v[Z] = w->v[Z];
}

__inline void VectorAdd(VECTOR *v, VECTOR *w) {
  v->v[X] = v->v[X] + w->v[X];
  v->v[Y] = v->v[Y] + w->v[Y];
  v->v[Z] = v->v[Z] + w->v[Z];
}

__inline void VectorSubtract(VECTOR *v, VECTOR *w) {
  v->v[X] = v->v[X] - w->v[X];
  v->v[Y] = v->v[Y] - w->v[Y];
  v->v[Z] = v->v[Z] - w->v[Z];
}

__inline void VectorMultiply(VECTOR *v, float scalar) {
  v->v[X] *= scalar;
  v->v[Y] *= scalar;
  v->v[Z] *= scalar;
}

__inline void VectorDivide(VECTOR *v, float scalar) {
  scalar = 1.0f / scalar;
  v->v[X] *= scalar;
  v->v[Y] *= scalar;
  v->v[Z] *= scalar;
}

__inline int VectorEqual(VECTOR *v, VECTOR *w) {
  if (v->v[X]==w->v[X] && v->v[Y]==w->v[Y] && v->v[Z]==w->v[Z]) return TRUE;
  return FALSE;
}

/* Cross Products are very important for things like lighting it
 * generates a scalar of length /v/ /w/ cos( theta )
 */
__inline float VectorDotProduct(VECTOR *v, VECTOR *w) {
  return ( v->v[X]*w->v[X] + v->v[Y]*w->v[Y] + v->v[Z]*w->v[Z] );
}

__inline float VectorLength(VECTOR *v) {
  return (float)sqrt(VectorDotProduct(v, v));
}

/* Cross Products are very important for things like lighting it
 * generates a vector v perpendicular to the plane defined by
 * vectors v1 and v2.
 *
 * Note that v is of length L(v1)*L(v2)*abs( sin( angle between v1 & v2 ) )
 * as a result of this you can use a cross product to find the angle between
 * two vectors, of course with all the sqrt() calls this would be very
 * expensive
 */
__inline void VectorCrossProduct(VECTOR *v, VECTOR *v1, VECTOR *v2) {
  float x;
  float y;
  float z;

  x = v1->v[Y] * v2->v[Z] - v1->v[Z] * v2->v[Y];
  y = v1->v[Z] * v2->v[X] - v1->v[X] * v2->v[Z];
  z = v1->v[X] * v2->v[Y] - v1->v[Y] * v2->v[X];

  v->v[X] = x;
  v->v[Y] = y;
  v->v[Z] = z;
}

__inline void VectorNormalize(VECTOR *v) {
  float length;

  length = 1.0f / VectorLength(v);
  v->v[X] = v->v[X] * length;
  v->v[Y] = v->v[Y] * length;
  v->v[Z] = v->v[Z] * length;
}

__inline float VectorGetYaw(VECTOR *v) {
  float l;
  float c;
  float yaw;

  l = (float)sqrt(v->v[X]*v->v[X] + v->v[Z]*v->v[Z]);
  if (l==0.0f) return 0.0f;
  if (v->v[Z]>0.0f) {
    c = (v->v[Z]) / l;
    if (v->v[X]<0.0f)
      yaw = 0.0f - DEGREE((float)acos(c));
    else
      yaw = DEGREE((float)acos(c));
  } else {
    c = (-1 * v->v[Z]) / l;
    if (v->v[X]<0.0f)
      yaw = 180.0f + DEGREE((float)acos(c));
    else
      yaw = 180.0f - DEGREE((float)acos(c));
  }
  return yaw;
}

__inline float VectorGetPitch(VECTOR *v) {
  float l;
  float c;
  float pitch;

  l = (float)sqrt(v->v[Y]*v->v[Y] + v->v[Z]*v->v[Z]);
  if (l==0.0f) return 0.0f;
  if (v->v[Z]>0.0f) {
    c = (v->v[Z]) / l;
    if (v->v[Y]<0.0f)
      pitch = 0.0f - DEGREE((float)acos(c));
    else
      pitch = DEGREE((float)acos(c));
  } else {
    c = (-1 * v->v[Z]) / l;
    if (v->v[Y]<0.0f)
      pitch = 180.0f + DEGREE((float)acos(c));
    else
      pitch = 180.0f - DEGREE((float)acos(c));
  }
  return pitch;
}

__inline float VectorGetRoll(VECTOR *v) {
  float l;
  float c;
  float roll;

  l = (float)sqrt(v->v[X]*v->v[X] + v->v[Y]*v->v[Y]);
  if (l==0.0f) return 0.0f;
  if (v->v[Y]>0.0f) {
    c = (v->v[Y]) / l;
    if (v->v[X]<0.0f)
      roll = 0.0f - DEGREE((float)acos(c));
    else
      roll = DEGREE((float)acos(c));
  } else {
    c = (-1 * v->v[Y]) / l;
    if (v->v[X]<0.0f)
      roll = 180.0f + DEGREE((float)acos(c));
    else
      roll = 180.0f - DEGREE((float)acos(c));
  }
  return roll;
}

//#ifdef UNTESTED_NEVER_USED
void VectorTransform(VECTOR *v, MATRIX *m) {
  VECTOR  w;
  float W;

  VectorCopy(&w, v);
  W       =  w.v[X]*m->m[3 ] + w.v[Y]*m->m[7 ] + w.v[Z]*m->m[11] + m->m[15];
  W       = 1.0f / W;
  v->v[X] = (w.v[X]*m->m[0 ] + w.v[Y]*m->m[4 ] + w.v[Z]*m->m[8 ] + m->m[12])*W;
  v->v[Y] = (w.v[X]*m->m[1 ] + w.v[Y]*m->m[5 ] + w.v[Z]*m->m[9 ] + m->m[13])*W;
  v->v[Z] = (w.v[X]*m->m[2 ] + w.v[Y]*m->m[6 ] + w.v[Z]*m->m[10] + m->m[14])*W;
}
//#endif
